<?php

/****************************************************************************

  Copyright (C) 2017-2020  Sendatel S.A.

  This file is part of TUIDD Wordpress SDK.
  TUIDD Wordpress SDK is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  TUIDD Wordpress SDK is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with TUIDD Wordpress SDK.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

require_once('TuiddException.php');
require_once( dirname(dirname(dirname(dirname(dirname(dirname( dirname( __FILE__ ))))))) . '/wp-load.php');

if(!isset($_SESSION)) 
{ 
    session_name('sdk-session');
    session_start();
}

class TuiddClient
{
	private $m_ClientSecret = '';
	private $m_ClientId = '';
	private $m_UrlRedirect = '';
	private $m_UrlDiscovery = '';
	private $m_UrlTuiddContent = '';
	private $m_UrlProveedoraToken = '';
	private $m_UrlUserInfoRedirect = '';

	private $m_AuthCode = '';
	private $m_accessTokenData = '';
	private $m_UserInfo = '';
	private $m_PaymentToken = '';

	private $m_LastServiceResult = '';
	private $m_UrlSendService = '';

	private $m_ContainerDiv = '';
	private $m_tuiddSDKPath = '';
	private $m_Scope = '';

	private $m_State= '';
    private $m_Nonce= '';
	private $m_Security = '';
	private $m_SaveUserInfo	= '';
    private $m_callback_function  = '';
    private $m_state_nonce_callback = '';
	private $m_ExecuteCallback = '';

	private $m_Language = '';
	private $m_HomePageUrl = '';

	private $m_init_time = '';
    private $m_Display = '';
    private $m_DEBUG = '';
	private $m_global_init = 0;

	function __construct()
	{
		$internal_conf = parse_ini_file( "internal_params.conf" );
		$this->m_UrlDiscovery = $internal_conf['tuidd_discovery'];
		$this->m_UrlTuiddContent = $internal_conf['tuidd_sdk'];

		$this->m_ClientId = get_option('tuidd_option_client_id');
		$this->m_Scope = get_option('tuidd_option_scope');
		$this->m_Security = get_option('tuidd_option_security_level');
		$this->m_ClientSecret = get_option( 'tuidd_option_secret' );
		$this->m_callback_function  = get_option( 'tuidd_option_callback_function' );
		$this->m_PaymentToken = get_option( 'tuidd_option_payment_token' );
		$this->m_UrlRedirect = get_option( 'tuidd_option_redirect_uri' );
		$this->m_HomePageUrl = get_option('tuidd_option_home_page_url');
		$this->m_UrlUserInfoRedirect = get_option('tuidd_option_home_login_url');
		$this->m_SaveUserInfo = get_option('tuidd_option_save_userinfo');
		$this->m_ExecuteCallback = get_option('tuidd_option_execute_callback');
		$this->m_state_nonce_callback = get_option('tuidd_option_state_nonce_callback');
        $this->m_DEBUG = get_option('tuidd_option_debug');
		$this->m_Display = 'page';
	}


	/**
	 * Gets Singleton's instance of TuiddClient
	 * @param  none
	 * @return [type]              [description]
	 */
	public static function getSingleton()
	{
		if ( !isset($_SESSION['tuidd']) || empty($_SESSION['tuidd']) )
		{
			return new self();
		}
		else
		{
            if( $_SESSION['tuidd'] instanceof TuiddClient ) {
                return $_SESSION['tuidd'];
            } else {
                $ret = new self();
                $ret = unserialize($_SESSION['tuidd']);
                return $ret;
            }
        }
	}


	private function updateContent() {
		$_SESSION['tuidd'] = serialize($this);
	}


	/**
     * LOGOUT and destroy session
     * @param  none
     * @return Bool True or Redirect
     */
    public function logout()
    {
        $home = $this->m_HomePageUrl;
        session_destroy('sdk-session');
        $_SESSION['tuidd'] = '';
        header("Location:" . $home );
    }


	public function discovery( )
	{
        $arrContextOptions=array(
            "ssl"=>array(
                "verify_peer"=>false,
                "verify_peer_name"=>false,
            ),
        );

		$mcg_url = $this->curl_file_get_contents( $this->m_UrlDiscovery . "/mcg_urls", false, stream_context_create($arrContextOptions));
		$urls_json = json_decode($mcg_url, true);

		$request_url = $urls_json['request_url'];
		$this->m_UrlProveedoraToken = (string) $urls_json['service_provider_url'];
		$this->m_UrlSendService = (string) $urls_json['send_service_url'];
		$this->updateContent();

		return $request_url;
	}

	//public bool getToken (TuiddGetTokenReq $object )
	public function getToken( $code )
	{
		$this->m_AuthCode = $code;
		$this->updateContent();

		$url = $this->m_UrlProveedoraToken;
		$headers = array();
        $headers[]     = 'Authorization: Basic '.base64_encode( $this->m_ClientId . ':' . $this->m_ClientSecret );
        $headers[]     = 'Content-Type: Application/x-www-form-urlencoded';

		$arrParams = array('grant_type' => 'authorization_code', 'code' => $this->m_AuthCode, 'redirect_uri' => $this->m_UrlRedirect );
		$params = http_build_query($arrParams);

        $result = $this->sendCurl($url, $params, $headers);
        if ( $this->m_DEBUG) {
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." ::  RESULT" . print_r($result,true));
        }
		$aux = (array)json_decode($result['response']);

		if ( empty($aux) || ! empty($aux['error']) )
		{
			if ( empty($aux))
			{
				throw new TuiddException("Error Durante el pedido del accessToken. respuesta vacio");
			}
			else
			{
				throw new TuiddException("Error Durante el pedido del accessToken. La descripcion del error es la siguiente: " . $aux['error_descripcion'] );
			}
		}
		else {
			$this->m_accessTokenData = $aux;
			$this->m_init_time = time();
			return true;
		}

	}

	// public bool getUserInfo (TuiddGetUserInfoReq $object )
	public function requestUserInfo()
	{
		$url = $this->m_UrlProveedoraToken;

		$headers = array('Content-Type: application/x-www-form-urlencoded', 'Authorization: Bearer ' . $this->m_accessTokenData['access_token']);
		$arrParams = array('grant_type' => 'authorization_code', 'code' => $this->m_AuthCode, 'redirect_uri' => $this->m_UrlRedirect);

		$params = http_build_query($arrParams);

		try {
			$result = $this->sendCurl($url, $params, $headers);
		}
		catch (Exception $e) {
	    	echo "Curl Error: ", $e->getMessage(), '\n';
		}

		$aux = (array)json_decode($result['response']);

		if ( empty($aux) || ! empty($aux['error_description']) )
		{
			if ( empty($aux))
			{
				throw new TuiddException("Error Durante el pedido del UserInfo. Respuesta vacia");
			}
			else
			{
				throw new TuiddException("Error Durante el pedido del UserInfo. La descripcion del error es la siguiente: " . $aux['error_description'] );
			}
		}
		else {
			$this->m_UserInfo = $aux;
			return true;
		}
	}

	/* Esta función sirve para procesar los datos del usuario.
     *  Que hacer con esos datos depende del poveedor.
     *  En este caso se realiza un redirect al servicio.
     */
	public function cfReceive()
	{
		$user_info = $this->m_UserInfo;
		$user_info['client_info']  = base64_encode( $this->m_ClientId . ':' . $this->m_ClientSecret );
		$user_info['access_token']  = $this->m_accessTokenData['access_token'];
		$user_info['id_token']  = $this->m_accessTokenData['id_token'];

		$this->m_UserInfo = $user_info;
		$this->updateContent();

		// Si quiere crear usuario
		// Si quiere ejecutar el callback
		// Redirifir a user
		if ($this->m_SaveUserInfo )
		{
			$this->tuidd_create_and_login_user();
		}
		if ($this->m_ExecuteCallback)
		{
			tuidd_action_callback_function();
		}
    	wp_redirect( $this->m_UrlUserInfoRedirect );
	}


	/**
	 * consulta si esta habilitado el debug
	 * @return boolean isDebugging
	 */	
	public function isDebugging()
	{
		return $this->m_DEBUG;
	}

	/**
     * Setting Nonce.
     * @param  string  nonce
     */
    public function setNonce ( $nonce )
    {
        $this->m_Nonce = $nonce;
        $this->updateContent();
    }


    /**
     * Setting State.
     * @param  string  state
     */
    public function setState ( $state )
    {
        $this->m_State = $state;
        $this->updateContent();
    }


    /**
     * Callback Function for consuming a service.
     * @param  array                        Service information     to be consumed
     * @return array            Status
     */
    public function consumeService ( $service )
    {
        $headers = array();
        $headers[]     = 'Authorization: Bearer '.base64_encode( $this->m_ClientId . ':' . $this->m_ClientSecret );
        $headers[]     = 'Content-Type: Application/json';

        $url = $this->m_UrlSendService;

        // Para Array multidimensionales
        $params = json_encode($service, JSON_FORCE_OBJECT);

        $result = $this->sendCurl($url, $params, $headers);

        $r = $result['response'];
        if ( $this->m_DEBUG)
        {
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." :: URL " . print_r($r,true));
        }

        if (strpos($r, "{" ) > 1)
        {
            $r = substr($r, strpos($r, "{" )-1, strlen($r));
        }

        $response = json_decode( $r, true);

        $st = 1;
        $statusMessage;

        if ($result['status'] == 200)
        {
            if ($response['transaction_id'] && $response['seq_num'] )
            {
                $statusMessage = 'OK';
                $rta = array( 'transaction_id' => $response['transaction_id'],
                              'seq_num' => $response['seq_num'],
                );
                $st = 0;
                if ($response['url_redirect'])
                {
                    $rta['url_redirect'] = $response['url_redirect'];
                }
            }
            else
            {
                $rta = null; //$response['error_data'];
                $statusMessage = 'Error. Respuesta Incompleta. Falta ';

                if (! $response['transaction_id'])
                {
                    $statusMessage = $statusMessage . 'el parametro transaction_id.';
                }
                else if (! $response['seq_num'])
                {
                    $statusMessage = $statusMessage . 'el parametro seq_num.';
                }
            }
        }
        else if ($result['status'] == 400)
        {
            $statusMessage = 'BAD REQUEST';
        }
        else if ($result['status'] == 500)
        {
            $statusMessage = 'INTERNAL SERVER ERROR';
        }
        else if ($result['status'] == 401)
        {
            $statusMessage = 'UNAUTHORIZED';
        }

        $serviceResponse = array( 'result' => $st,
                                  'status' => array('description' => $statusMessage,
                                                    'code' => $result['status']),
                                  'response' => $rta,
        );

        return $serviceResponse;
    }


    /**
     * Get Access TOKEN.
     * @param  array Service information to be consumed
     * @return array Status
     */
    public function getAccessToken() {
        return $this->m_accessTokenData['access_token'];
    }

    /**
     * Get Survey service TOKEN.
     * @param  array Service information to be consumed
     * @return array Status
     */
    public function getSurveyServiceToken() {
        return $this->m_accessTokenData['tuidd.com.survey_token'];
    }

    /**
     * Get Advertisement service TOKEN.
     * @param  array Service information to be consumed
     * @return array Status
     */
    public function getAdvertisingServiceToken() {
        return $this->m_accessTokenData['tuidd.com.advertising_token'];
    }


    /**
     * Get Notification service TOKEN.
     * @param  array Service information to be consumed
     * @return array Status
     */
    public function getNotificationServiceToken()
    {
        return $this->m_accessTokenData['tuidd.com.notification_token'];
    }


    /**
	 * Config param getter
	 * @return client_id
	 */
	public function getClientID()
    {
		return $this->m_ClientId;
	}


	/**
	 * Config param getter
	 * @return scope
	 */
	public function getScope()
    {
		return $this->m_Scope;
	}


	/**
	 * Config param getter
	 * @return security_level
	 */
	public function getSecurityLevel()
    {
		return $this->m_Security;
	}


	/**
	 * Config param getter
	 * @return redirect_uri
	 */
	public function getRedirectURI()
    {
		return $this->m_UrlRedirect;
	}


	/**
	 * Config param getter
	 * @return default_language
	 */
	public function getLanguage()
    {
		return $this->m_Language;
	}

	/**
	 * Config param getter
	 * @return container_div
	 */
	public function getContainerDiv()
    {
		return $this->m_ContainerDiv;
	}

	/**
	 * Config param getter
	 * @return tuidd_sdk_path
	 */
	public function getTuiddSDKPath()
    {
		return $this->m_tuiddSDKPath;
	}

	/**
	 * Config param getter
	 * @return home_page_url
	 */
	public function getHomePageURL()
    {
		return $this->m_HomePageUrl;
	}


	/**
	 * Config param getter
	 * @return home_page_url
	 */
	public function getDiscoveryUrl()
    {
		return $this->m_UrlDiscovery;
	}


	/**
	 * If Discovery URLS had been set
	 * @return Bool
	 */
	public function hasDiscoveryUrls()
    {
		return ( isset($this->m_UrlProveedoraToken) && $this->m_UrlProveedoraToken != '');
	}


	/**
	 * Config param getter
	 * @return url_tuidd_content
	 */
	public function getTuiddContentURL()
    {
		return $this->m_UrlTuiddContent;
	}


	/**
	 * Simply packs User information and delivers it 
	 * @return array  userInfo
	 */
	public function getUserInfo()
    {
		$userInfo = $this->m_UserInfo;
		return $userInfo;
	}


    public function getStateNonceCallback()
    {
        return $this->m_state_nonce_callback;
    }


	/**
	 * Get Survey service TOKEN. 
	 * @param  array 			Service information	to be consumed
	 * @return array            Status
	 */
	public function generateStateAndNonce()
    {
        // Se llama a la funcion de callback establecida en archivo de config
        // para generar state y nonce
        if ( $this->getStateNonceCallback() != '' )
        {
            //require_once( $this->getCallbackFunctionFile() );
            call_user_func( $this->getStateNonceCallback() );
        }
	}


    public function getState()
    {
        return $this->m_State;
    }


	/**
	 * Genera un state que contiene el state del sitio y un state interno del
	 * sdk.
	 * 
	 * @param  string $clientState el state que notifico el sitio.
	 * @param  string $sdkState    state con informaicon util para el sdk.
	 * @return string              json con los states en base64_url.
	 */
	public function generateSdkState($clientState, $sdkState='')
	{
		$jsonState = json_encode([
			'client_state' => $clientState,
			'sdk_state' => $sdkState
        ]);
		if ($this->__DEBUG__)
		{
			syslog(LOG_INFO, "TUIDD SDK :: ". __METHOD__ ." :: state: $jsonState");
		}

		return self::base64urlEncode($jsonState);
	}


	/**
	 * deserializa el rawState devolciendo un array asociativo con los campos:
	 * [ 'client_state' => $clientState, 'sdk_state' => $sdkState ]
	 * 
	 * @param  String $rawState json con los states en base64_url.
	 * @return array            Array con 2 campos client_state y action_state.
	 */
	public function readSdkState($rawState)
	{
		$jsonState = self::base64urlDecode($rawState);

		if ($this->__DEBUG__)
		{
			syslog(LOG_INFO, "TUIDD SDK :: ". __METHOD__ ." :: state: $jsonState");
		}
		$ret = json_decode($jsonState, TRUE);
		if ($this->__DEBUG__)
		{
			syslog(LOG_INFO, "TUIDD SDK :: ". __METHOD__ ." :: return: ".print_r($ret,TRUE));
		}
		return $ret;
	}


    public function getNonce()
    {
        return $this->m_Nonce;
    }


	/**
	 * validateLogin Validates id_token
	 * @return true  if valid
	 */
	public function validateLogin()
	{
		$parent = dirname( dirname( dirname( __FILE__ ) ) );
		require_once( 'lib/jwtManager/JWT.php' );

		$id_token = $this->m_accessTokenData['id_token'];
		$cs = $this->m_ClientSecret;

		try
		{
			$decoded = JWT::decode($id_token, $cs, array('HS256'));
			$decoded = (array) $decoded;
			return true;
		}
		catch(Exception $e)
		{
			echo 'Message: ' .$e->getMessage();
		}
	}


	/* Esta funcion es la que esta encargada de hacer los envios por curl y transmitir la informacion
     */
	private	function sendCurl($url,$params, $headers)
	{
		$params = urldecode($params);

        if ( $this->m_DEBUG) {
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." :: URL " . print_r($url,true));
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." :: PARAMS " . print_r($params ,true));
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." :: HEADERS " . print_r($headers,true));
        }

		$ch = curl_init();

		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_HTTPGET, FALSE);
        curl_setopt($ch, CURLOPT_POSTFIELDS, $params);

        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, false);

        curl_setopt($ch, CURLOPT_HEADER, true);
        curl_setopt($ch, CURLOPT_HTTPHEADER,  $headers);

        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);

        $response = curl_exec($ch);
        if ( $this->m_DEBUG) {
            syslog(LOG_INFO, "TUIDD SDK :: ". basename(__FILE__, ".php") ." ::  CURL RESPONSE " . print_r($response,true));
        }
		$http_status     = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $header_size     = curl_getinfo($ch, CURLINFO_HEADER_SIZE);
        $responseBody    = substr($response, $header_size);

        $r = json_decode( $responseBody );

        if ( curl_error($ch) )
        {
            curl_close($ch);
            throw new Exception("Error Durante el envio de Curl");
        }
        curl_close($ch);

		return array(
            'status' => $http_status,
            'response' => $responseBody
        );

	}


    private function curl_file_get_contents($file)
    {
        if (function_exists('curl_version'))
        {
            $curl = curl_init();
            curl_setopt($curl, CURLOPT_URL, $file);
            curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
            curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, 0);
            curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 0);

            $content = curl_exec($curl);
            curl_close($curl);
        }
        else if (file_get_contents(__FILE__) && ini_get('allow_url_fopen'))
        {
            $arrContextOptions=array(
                "ssl"=>array(
                    "verify_peer"=>false,
                    "verify_peer_name"=>false,
                ),
            ); 
            $content = file_get_contents($file, false, stream_context_create($arrContextOptions));
        }
        else
        {
            echo 'You have neither cUrl installed nor allow_url_fopen activated. Please setup one of those!';
        }
        return $content;
    }


	public static function base64urlEncode($str)
	{
		return rtrim( strtr( base64_encode( $str ), '+/', '-_'), '=');
	}


	public static function base64urlDecode($str)
	{
		return base64_decode( strtr( $str, '-_', '+/') . str_repeat('=', 3 - ( 3 + strlen( $str )) % 4 ));
	}


    private function tuidd_create_and_login_user()
    {
    	$create_user = true;
    	$user_info = $this->getUserInfo();

    	$tuidd_user_mail = $this->tuidd_user_has_mail();
    	$tuidd_user_name = $this->tuidd_user_has_name();
    	$tuidd_user_sub = $this->tuidd_user_get_sub();

    	// Trata de logearlo
    	if ($tuidd_user_mail)
    	{
    		$usere_id = email_exists( $tuidd_user_mail );
    		if ($usere_id)
    		{
    			$this->tuidd_update_wp_user($usere_id);
    			$create_user = false;
    		}
    	}
    	else if ($tuidd_user_name)
    	{
    		$usern_id = username_exists( $tuidd_user_name );
    		$meta = get_user_meta( $usern_id, 'tuidd');

    		if ($usern_id && $tuidd_user_sub == $meta['sub'])
    		{
    			$this->tuidd_update_wp_user($usern_id);
    			$create_user = false;
    		}
    	}
    	else if ($tuidd_user_sub)
    	{
    		$users_id = username_exists( $tuidd_user_sub );
    		if ($users_id) {
    			$this->tuidd_update_wp_user($users_id);
    			$create_user = false;
    		}
        }
    	// Fallo todo se crea nuevo usuario
    	if ($create_user) {

			if ($tuidd_user_name)
			{
				$user_name = $tuidd_user_name;
			}
			else
			{
				$user_name = $tuidd_user_sub;
			}
			$user_email = $tuidd_user_mail;

			$password = wp_generate_password( 12, false );
			$user_id = wp_create_user( $user_name, $password, $user_email );

			// Set the role
			$user = new WP_User( $user_id );
			$user->set_role( 'contributor' );
			wp_set_current_user( $user_id );

			$this->tuidd_update_wp_user($users_id);
    	}
	}


	private function tuidd_user_has_mail()
	{
		return $this->m_UserInfo['mail'];
	}


	private function tuidd_user_has_name()
	{
		return $this->m_UserInfo['name'];
	}


	private function tuidd_user_get_sub()
	{
		return $this->m_UserInfo['sub'];
	}


	private function tuidd_update_wp_user($user_id)
	{
		add_user_meta($user_id, 'tuidd', $this->m_UserInfo, true);
		return	;
	}
}
?>
